@@ -15,10 +15,11 @@ end module global
 program namelist_use
   use global
   real    :: rrr
-  namelist /nml2/ ii, rrr    ! Concatenate use and host associated variables.
+! Concatenating use and host associated variables is an extension.
+  namelist /nml2/ ii, rrr ! { dg-warning "already is USE associated" }
   open (10, status="scratch")
-  write (10,*) "&NML1 aa=lmno ii=1 rr=2.5 /"
-  write (10,*) "&NML2 aa=pqrs ii=2 rrr=3.5 /"
+  write (10,*) "&NML1 aa='lmno' ii=1 rr=2.5 /"
+  write (10,*) "&NML2 aa='pqrs' ii=2 rrr=3.5 /"
   rewind (10)
   read (10,nml=nml1,iostat=i)
   if ((i.ne.0).or.(aa.ne."lmno").or.(ii.ne.1).or.(rr.ne.2.5)) call abort ()
@@ -28,5 +29,3 @@ program namelist_use
 
   close (10)
 end program namelist_use
-
-! { dg-final { cleanup-modules "global" } }
